import logging

from key_spyder.defaults import DEFAULT_PATH, NOW


class Logger:
    def __init__(self, name, verbose=False):
        fh_path = f"{DEFAULT_PATH}/logs/{name}_{NOW}.log"
        file_handler = logging.FileHandler(fh_path, "w", encoding="utf-8")
        logging.basicConfig(
            format="%(asctime)s %(levelname)s %(message)s",
            level=logging.INFO,
            encoding="utf-8"
        )
        self.logger = logging.getLogger(name)
        self.logger.addHandler(file_handler)
        if verbose:
            self.logger.setLevel(10)

    def info(self, message):
        self.logger.info(message)

    def error(self, message):
        self.logger.error(message)

    def warning(self, message):
        self.logger.warning(message)

    def critical(self, message):
        self.logger.critical(message)

    def debug(self, message):
        self.logger.debug(message)

    def exception(self, message):
        self.logger.exception(message)

    def log(self, level, message):
        self.logger.log(level, message)

    def set_level(self, level):
        self.logger.setLevel(level)



from importlib import resources
from os import makedirs, path
from urllib.parse import urljoin, urlparse

from bs4 import BeautifulSoup
from key_spyder.logs import Logger
from key_spyder.defaults import DEFAULT_PATH, NOW
from requests.exceptions import RequestException
from requests_cache import CachedSession
from symspellpy import SymSpell


class Crawler:
    def __init__(self,
                 url: str,
                 params: dict = None,
                 keywords: list[str] = None,
                 spell_check: bool = False,
                 recursive: bool = False,
                 output_directory: str = None,
                 verbose: bool = False,
                 clear_cache: bool = False,
                 known_urls: list[str] = None):

        self.logger = Logger(__class__.__name__, verbose)
        if params is None:
            params = {}
        if keywords is None:
            keywords = []
        if output_directory is None:
            output_directory = DEFAULT_PATH

        for folder in ["logs", "results"]:
            dir_path = path.join(output_directory, folder)
            if not path.exists(dir_path):
                makedirs(dir_path)

        self.first_url = url
        self.urls_to_visit = [self.first_url]
        self.keywords = keywords
        self.spell_check = spell_check
        self.params = params
        self.recursive = recursive
        self.output_directory = output_directory

        self.visited_urls = []
        self.results = ["url,params,field_0,field_1,type\n"]

        if known_urls:
            self.urls_to_visit = known_urls

        cache_name = f"{output_directory}/cache"
        self.session = CachedSession(cache_name, expire_after=86400)
        if clear_cache:
            self.session.cache.clear()

    @property
    def all_urls(self):
        return self.urls_to_visit + self.visited_urls

    def get_html(self, url):
        self.logger.debug(f"Getting: {url} {f'with {self.params}' if self.params else ''}")
        try:
            response = self.session.get(url, params=self.params, allow_redirects=False)
            if response.from_cache:
                self.logger.debug(f"Got: {url} from cache")
            else:
                self.logger.debug(f"Got: {url} in {response.elapsed.total_seconds()} seconds")
        except RequestException as e:
            self.logger.exception(e)
        else:
            return response.text

    @staticmethod
    def get_links(url, html):
        """
        For a given url, search all anchor tags, return a list of all new links.
        """
        soup = BeautifulSoup(html, 'html.parser')
        parsed_url = urlparse(url)
        prot_host_tld = f"{parsed_url.scheme}://{parsed_url.hostname}"

        # Find all anchor tags on the page.
        for link in soup.find_all('a'):
            path = link.get('href')
            # Make sure the href was set.
            if path:
                # If it's an internal link, prepend the hostname to the path.
                if path.startswith('/'):
                    path = urljoin(prot_host_tld, path)
                # If the new path is still on the beginning host.
                if prot_host_tld in path:
                    yield path

    def get_keywords(self, url, html):
        """
        For a given url, check for keywords and write them to the results list.
        """
        soup = BeautifulSoup(html, 'html.parser')
        text = soup.body.get_text().split("\n")

        text = [line.strip() for line in text if line.strip()]

        for line in text:
            for keyword in self.keywords:
                self.logger.debug(f"Checking: '{keyword}' in '{line}'")
                if keyword.lower() in line.lower():
                    self.logger.info(f"Found: '{keyword}' in '{line}' on {url}")
                    self.write_line(url, keyword, line, "keyword")

    def check_spellings(self, url, html):
        sym_spell = SymSpell(max_dictionary_edit_distance=2, prefix_length=7)
        dict_path = resources.path("key_spyder", "frequency_dictionary_en_82_765.txt")
        sym_spell.load_dictionary(dict_path, 0)

        soup = BeautifulSoup(html, 'html.parser')
        text = soup.body.get_text().split("\n")

        text = [line.strip() for line in text if line.strip()]

        for line in text:
            suggestions = sym_spell.lookup_compound(line, max_edit_distance=2)
            for suggestion in suggestions:
                self.logger.info(f"Suggested Correction: '{suggestion.term}' for '{line}' on {url}")
                self.write_line(url, suggestion.term, line, "suggestion")

    def crawl(self, url, html):
        """
        For a given url, Discover new links on that page.
        """
        self.logger.info(f'Crawling: {url}')
        for link in self.get_links(url, html):
            if link not in self.all_urls:
                self.logger.info(f'Discovered: {link}')
                self.urls_to_visit.append(link)

    def write_line(self, url, field_0, field_1, result_type):
        self.results = self.results + [f'{url},{self.params},{field_0},"{field_1}",{result_type}\n']

    def write_results(self):
        filename = f"{urlparse(self.first_url).netloc}_results_{NOW}.csv"
        filepath = path.join(self.output_directory, "results", filename)
        if len(self.results) > 1:
            with open(filepath, "w") as f:
                f.writelines(self.results)
        else:
            self.logger.info(f"No results.")

    def run(self):
        while self.urls_to_visit:
            url = self.urls_to_visit.pop(0)
            self.visited_urls.append(url)
            html = self.get_html(url)
            if html:
                if self.recursive:
                    self.crawl(url, html)
                if self.keywords:
                    self.get_keywords(url, html)
                if self.spell_check:
                    self.check_spellings(url, html)
        self.write_results()
        self.logger.info(f"Output can be found at: {self.output_directory}")

    def __exit__(self):
        self.write_results()
        self.session.close()



from argparse import ArgumentError, ArgumentParser
from logging import INFO
from multiprocessing import Process, log_to_stderr
from os import makedirs, path

from key_spyder.crawler import Crawler
from key_spyder.defaults import DEFAULT_PATH
from key_spyder.sitemap import Sitemapper

parser = ArgumentParser(
    prog='key-spyder',
    description='Crawl websites for keywords')
parser.add_argument(
    '-u', '--urls',
    action='store',
    dest='urls',
    nargs='+',
    type=str,
    required=True,
    help='Entrypoint URLs to begin crawling',
    metavar='URLS'
)
parser.add_argument(
    '-p', '--params',
    action='store',
    dest='params',
    nargs='+',
    required=False,
    help='Parameters for requests while crawling',
    metavar='PARAMS'
)
parser.add_argument(
    '-k', '--keywords',
    action='store',
    dest='keywords',
    nargs='+',
    required=False,
    help='Keywords to search for in crawled pages',
    metavar='KEYWORDS'
)
parser.add_argument(
    '-r', '--recursive',
    action='store_true',
    dest='recursive',
    required=False,
    default=False,
    help='Recursively crawl linked pages'
)
parser.add_argument(
    '-o', '--output',
    action='store',
    dest='output',
    help='Output directory',
    default=None,
    metavar='OUTPUT'
)
parser.add_argument(
    '-v', '--verbose',
    action='store_true',
    dest='verbose',
    required=False,
    default=False,
    help='Verbose output'
)
parser.add_argument(
    '-c', '--clear-cache',
    action='store_true',
    dest='clear_cache',
    required=False,
    default=False,
    help='Clear cache before crawling'
)
parser.add_argument(
    '-s', '--sitemap',
    action='store_true',
    dest='sitemap',
    required=False,
    default=False,
    help='Use Sitemap for faster crawling'
)
parser.add_argument(
    '-sc', '--spell-check',
    action='store_true',
    dest='spell_check',
    required=False,
    default=False,
    help='Spell check crawled pages'
)


def run_from_cli(url, params, keywords, spell_check, recursive, output, verbose, clear_cache,
                 known_urls):
    Crawler(
        url=url,
        params=params,
        keywords=keywords,
        spell_check=spell_check,
        recursive=recursive,
        output_directory=output,
        verbose=verbose,
        clear_cache=clear_cache,
        known_urls=known_urls
    ).run()


# Main function
def main():
    def parse_params(params_args):
        for param in params_args:
            if '=' in param:
                key, value = param.split('=')
                yield key, value
            else:
                raise ArgumentError(
                    f"Invalid parameter argument: {param}, parameters must be in the form 'key=value'")

    args = parser.parse_args()

    output_path = DEFAULT_PATH
    if args.output:
        if not path.exists(args.output):
            makedirs(args.output)
            output_path = args.output

    sitemaps = {}
    if args.sitemap:
        sitemaps = {url: Sitemapper(url) for url in args.urls}
        for sitemap in sitemaps.values():
            sitemap.to_csv(output_path)

    params = dict(parse_params(args.params)) if args.params else None

    processes = []
    for url in args.urls:
        known_urls = list(sitemaps[url].all_urls["loc"]) if sitemaps else []
        processes.append(Process(name=url,
                                 target=run_from_cli,
                                 args=(url, params, args.keywords,
                                       args.spell_check, args.recursive,
                                       output_path, args.verbose,
                                       args.clear_cache, known_urls)))

    if len(processes):
        if args.verbose:
            log_to_stderr(INFO)
        [process.start() for process in processes]
        [process.join() for process in processes]



from os import makedirs, path
from urllib import request, robotparser
from urllib.parse import urlparse

import pandas as pd
from bs4 import BeautifulSoup
from key_spyder.logs import Logger
from key_spyder.defaults import NOW


class Sitemapper:
    def __init__(self, url, verbose=False):
        self.logger = Logger(__class__.__name__, verbose)
        self.url = urlparse(url)

        self.__filename = f"{self.url.netloc}_sitemap_{NOW}.csv"
        self.__sitemap_index = self.__get_sitemap_url()
        self.all_urls = self.__get_all_urls(self.__sitemap_index)

    def __get_sitemap_url(self):
        robots = f"{self.url.scheme}://{self.url.netloc}/robots.txt"
        rp = robotparser.RobotFileParser()
        rp.set_url(robots)
        rp.read()
        sitemaps = rp.site_maps()
        if sitemaps:
            self.logger.debug(f"Found sitemap index: {sitemaps[0]}")
            return sitemaps[0]
        else:
            self.logger.debug(f"No sitemap index found")
            return None

    @staticmethod
    def __get_sitemap(url):
        r = request.urlopen(url)
        xml = BeautifulSoup(r, 'lxml-xml', from_encoding=r.info().get_param('charset'))
        return xml

    @staticmethod
    def __get_sitemap_type(xml):
        sitemapindex = xml.find('sitemapindex')
        sitemap = xml.find('sitemap')
        if sitemapindex:
            return 'sitemapindex'
        elif sitemap:
            return 'sitemap'
        else:
            return None

    @staticmethod
    def __sitemap_to_df(xml, name=None):
        df = pd.DataFrame(
            columns=['loc', 'changefreq', 'priority', 'domain', 'sitemap_name'])
        urls = xml.find_all("url")
        for url in urls:
            if xml.find("loc"):
                loc = url.findNext("loc").text
                domain = f'{urlparse(loc).netloc}'
            else:
                loc = domain = ''

            changefreq = url.findNext("changefreq").text if xml.find("changefreq") else ''
            priority = url.findNext("priority").text if xml.find('priority') else ''
            sitemap_name = name if name else ''

            df.loc[len(df)] = [loc, changefreq, priority, domain, sitemap_name]
        return df

    @staticmethod
    def __get_child_sitemaps(xml):
        sitemaps = xml.find_all('sitemap')
        return [sitemap.findNext('loc').text for sitemap in sitemaps]

    def __get_all_urls(self, url):
        xml = self.__get_sitemap(url)
        sitemap_type = self.__get_sitemap_type(xml)

        sitemaps = self.__get_child_sitemaps(xml) if sitemap_type == 'sitemapindex' else [url]

        df = pd.DataFrame(columns=['loc', 'changefreq', 'priority', 'domain', 'sitemap_name'])
        for sitemap in sitemaps:
            sitemap_xml = self.__get_sitemap(sitemap)
            df_sitemap = self.__sitemap_to_df(sitemap_xml, name=sitemap)
            df = pd.concat([df, df_sitemap], ignore_index=True)
        self.logger.info(f"Found {len(df)} URLs for {self.__sitemap_index}")
        return df

    def to_csv(self, filepath):
        sitemap_path = path.join(filepath, "sitemaps")
        if not path.exists(sitemap_path):
            makedirs(sitemap_path)
        self.logger.info(f"Saving sitemap to {sitemap_path}")
        self.all_urls.to_csv(f"{sitemap_path}/{self.__filename}", index=False)
